# flask imports
from flask_login import LoginManager, login_user, logout_user, login_required, current_user
from flask import Blueprint, request, render_template, redirect, url_for, flash, abort, jsonify
from flask_login import login_user, logout_user, login_required, current_user
from flask_wtf import FlaskForm
from wtforms import StringField, EmailField, PasswordField, SubmitField, SelectField, BooleanField
from wtforms.validators import InputRequired, DataRequired, Optional, Email, ValidationError
from werkzeug.security import generate_password_hash, check_password_hash
from http import HTTPStatus
from flask_iam.models import IAModels
from flask_iam.utils import root_required, role_required

class IAM:
    class LoginForm(FlaskForm):
        username = StringField('Username', validators=[InputRequired()])
        password = PasswordField('Password', validators=[InputRequired()])
        submit = SubmitField()

    class RegistrationForm(FlaskForm):
        username = StringField('Username', validators=[InputRequired()])
        email = EmailField('Email', validators=[InputRequired(), Email()], description='''
            Email is only used for site administration, and not shared with 3rd parties.
        ''')
        password = PasswordField('Password', validators=[InputRequired()])
        submit = SubmitField()

    class RoleForm(FlaskForm):
        name = StringField('Role name', validators=[InputRequired()])
        submit = SubmitField()

    class RoleAssignForm(FlaskForm):
        user_id = SelectField('User')
        role_id = SelectField('Role')
        submit = SubmitField()

    class ProfileForm(FlaskForm):
        username = StringField('Username', validators=[InputRequired()])
        email = EmailField('Email', validators=[InputRequired(), Email()], description='''
            Email is only used for site administration
            and confirmation of orders.
        ''')
        phone = StringField('Phone', validators=[Optional()],
            description='Optional, if you prefer being contacted by phone.'
        )
        address = StringField('Address', validators=[Optional()],
            description='Optional, if you require deliverable goods from here.'
        )
        old_password = PasswordField('Current password', validators=[Optional()])
        new_password = PasswordField('New password', validators=[Optional()])
        submit = SubmitField('Update profile')

        def validate_phone(self, phone):
            import phonenumbers
            try:
                p = phonenumbers.parse(phone.data)
                if not phonenumbers.is_valid_number(p):
                    raise ValueError()
            except (phonenumbers.phonenumberutil.NumberParseException, ValueError):
                raise ValidationError('Invalid phone number. Provide with country code, e.g. "+32..."')

        def validate_old_password(self, old_password):
            if not check_password_hash(current_user.password_hash, old_password.data):
                raise ValidationError('This is not the user you are looking for')

    def __init__(self, db, app=None, url_prefix='/auth', root_role='admin'):
        self.db = db
        self.url_prefix = url_prefix
        self.root_role = root_role
        self.login_manager = LoginManager()
        self.models = IAModels(db)

        @self.login_manager.user_loader
        def load_user(user_id):
            return self.models.User.query.get_or_404(user_id)
        self.login_manager.login_view = 'iam_blueprint.login'
        self.blueprint = Blueprint(
            'iam_blueprint', __name__,
            url_prefix=self.url_prefix,
            template_folder='templates'
        )

        self.blueprint.add_url_rule("/", 'iam', view_func=self.iam_index, methods=['GET'])
        self.blueprint.add_url_rule("/user/add", 'register', view_func=self.add_user_form, methods=['GET','POST'])
        self.blueprint.add_url_rule("/api/user/add", 'api_register', view_func=self.add_user_api, methods=['POST'])
        self.blueprint.add_url_rule("/user/login", 'login', view_func=self.login_user_form, methods=['GET','POST'])
        self.blueprint.add_url_rule("/user/logout", 'logout', view_func=self.logout_user, methods=['GET','POST'])
        self.blueprint.add_url_rule("/api/user/login", 'api_login', view_func=self.login_user_api, methods=['GET','POST'])
        self.blueprint.add_url_rule("/role/add", 'add_role', view_func=self.add_role, methods=['GET','POST'])
        self.blueprint.add_url_rule("/role/assign", 'assign_role', view_func=self.assign_role, methods=['GET','POST'])
        self.blueprint.add_url_rule("/admin", 'admin' , view_func=self.admin, methods=['GET'])
        self.blueprint.add_url_rule("/api/key/get", 'get_api_key', view_func=self.get_api_key, methods=['GET','POST'])

    def init_app(self, app):
        app.extensions['IAM'] = self # link for decorator access
        self.login_manager.init_app(app)
        app.register_blueprint(
            self.blueprint, url_prefix=self.url_prefix
        )
        # Set menu
        if 'fefset' in app.extensions:
            fef = app.extensions['fefset']
            fef.settings['side_menu_name_function'] = self.side_menu_name #'Account'
            fef.add_side_menu_entry('Login', f"{self.url_prefix}/user/login", role=None)#url_for('iam_blueprint.login'))        
            fef.add_side_menu_entry('Register', f"{self.url_prefix}/user/add", role=None)#url_for('iam_blueprint.register'))
            fef.add_side_menu_entry('Logout', f"{self.url_prefix}/user/logout", role=True)
            fef.add_side_menu_entry('Admin', f"{self.url_prefix}/", role=self.root_role)
            self.headless = False
        else:
            app.logger.warning(
                'No frontend available, operating in headless mode.'
                'If this is unintended, be sure to init "fefset" before "IAM" extension'
            )
            self.headless = True

    def side_menu_name(self):
        return f"Hi, {current_user}" if current_user.is_authenticated else 'Account'

    @root_required
    def iam_index(self):
        return render_template('IAM/index.html')

    def add_user(self, username, email, password):
        first_user = not bool(self.models.User.query.all())
        new_user = self.models.User(
            username=username,
            email=email,
            password_hash=generate_password_hash(password),
            role=self.root_role if first_user else 'viewer',
            enabled=first_user #if too many users disable
        )
        
        self.db.session.add(new_user)
        self.db.session.commit()
    
    def add_user_form(self):
        form = self.RegistrationForm()
        if form.validate_on_submit():
            self.add_user(
                username=form.username.data,
                email=form.email.data,
                password=form.password.data,
            )
            flash("User was created")

            return render_template('IAM/registration_success.html') #redirect('/')
        return render_template('uxfab/form.html', form=form, title='Register')
    
    def add_user_api(self):
        """
        API call route to add users.
        First user gets admin rights.

        Example:
          >>> import requests
          >>> r = requests.post(
          ...   'http://localhost:5000/auth/api/user/add',
          ...   json={'username':'admin',
          ...   'email':'admin@admin.be','password':'admin'}
          ... )
        
        """
        data = request.get_json()  # Get JSON data from the API call
        
        if data:  # Ensure data is provided
            # Check if the required fields are present in the JSON
            if 'username' in data and 'email' in data and 'password' in data:
                self.add_user(
                    username=data['username'],
                    email=data['email'],
                    password=data['password']
                )
                
                # Return a JSON success message
                return jsonify({"message": "User was created successfully", "status": "success"}), 201
            else:
                # Return an error message if fields are missing
                return jsonify({"message": "Missing required fields", "status": "error"}), 400
        else:
            # Return an error message if no JSON data is provided
            return jsonify({"message": "Invalid request: No JSON data", "status": "error"}), 400
    
    @root_required
    def add_role(self):
        form = self.RoleForm()
        if form.validate_on_submit():
            new_role = self.models.Role()
            form.populate_obj(new_role)
            self.db.session.add(new_role)
            self.db.session.commit()

            flash("Role was created")

            return redirect(url_for('iam_blueprint.iam'))
        return render_template('uxfab/form.html', form=form, title='Add role')

    @root_required
    def assign_role(self):
        form = self.RoleAssignForm()
        form.role_id.choices = [(r.id, r) for r in self.models.Role.query.all()]
        form.user_id.choices = [(r.id, r) for r in self.models.User.query.all()]
        if form.validate_on_submit():
            new_role_registration = self.models.RoleRegistration()
            form.populate_obj(new_role_registration)
            self.db.session.add(new_role_registration)
            self.db.session.commit()

            flash("Role was assigned")

            return redirect(url_for('iam_blueprint.iam'))
        return render_template('uxfab/form.html', form=form, title='Assign role')

    def login_user(self, username, password):
        user = self.models.User.query.filter_by(username=username).first()
        if user:
            if check_password_hash(user.password_hash, password):
                login_user(user)#, remember=form.remember.data)
                return True
        abort(404)

    def login_with_key(self, api_key):
        key = self.models.APIKey.query.filter_by(key=api_key).first()
        if key:
            login_user(self.models.User.query.get_or_404(key.user_id))
            return True
        abort(404)
        
    def login_user_form(self):
        form = self.LoginForm()
        if form.validate_on_submit():
            # Login and validate the user.
            # user should be an instance of your `User` class
            user_login_success = self.login_user(
                username = form.username.data,
                password = form.password.data
            )
            flash('Logged in successfully.')
            next = request.args.get('next')
            # is_safe_url should check if the url is safe for redirects.
            # See http://flask.pocoo.org/snippets/62/ for an example.
            #if not is_safe_url(next):
            #    return flask.abort(400)
            return redirect(next or '/')
        return render_template('IAM/login.html', form=form, title='Login')

    def logout_user(self):
        logout_user()
        return redirect('/')

    def login_user_api(self):
        """
        Important to work with sessions.
        Example login with key, but user/password works also.

        Example:
          >>> import requests
          >>> session = requests.Session()
          >>> session.post(
          ...   'http://localhost:5000/auth/api/user/login',
          ...   json={'key':key}
          ...   #or: json={'username':...,'password':...}
          ... )

        Continue with this session for your `post` or `get` calls.
        """
        data = request.get_json()
        
        if data:  # Ensure data is provided
            # Check if the required fields are present in the JSON
            if 'username' in data and 'password' in data:
                user_login_success = self.login_user(
                    username=data['username'],
                    password=data['password']
                )
                
                # Return a JSON success message
                return jsonify({"message": "User was logged in successfully", "status": "success"}), 201
            elif 'key' in data:
                user_login_sucess = self.login_with_key(data['key'])
                # Return a JSON success message
                return jsonify({"message": "User was logged in successfully", "status": "success"}), 201
            else:
                # Return an error message if fields are missing
                return jsonify({"message": "Missing required fields", "status": "error"}), 400
        else:
            # Return an error message if no JSON data is provided
            return jsonify({"message": "Invalid request: No JSON data", "status": "error"}), 400
        
    @login_required
    def get_api_key(self):
        """
        Example:
          >>> r = requests.post(
          ...   'http://localhost:5000/auth/api/user/add',
          ...   json={'username':'admin',
          ...   'email':'admin@admin.be','password':'admin'}
          ... )                                   
          >>> s = requests.Session()
          >>> r = s.post(
          ...   'http://localhost:5000/auth/api/user/login',
          ... json={'username':'admin','password':'admin'}
          ... )
          >>> r = s.post('http://localhost:5000/auth/api/key/get',json={})
          >>> print(r.json()) # shows API key
        """
        data = request.get_json()

        expiration = None
        if data:  # Ensure data is provided
            # Check if the required fields are present in the JSON
            if 'expiration' in data:
                expiration = data['expiration']
        
        key = self.models.APIKey(
            user_id = current_user.id,
            expiration = expiration
        )
        self.db.session.add(key)
        self.db.session.commit()
        return jsonify({"key": key.key, "status": "success"}), 201

    # Profile page
    @login_required
    def profile(self):
        form = self.ProfileForm(obj=current_user)
        if form.validate_on_submit():
            form.populate_obj(current_user)
            if form.old_password.data and form.new_password.data:
                current_user.password_hash = generate_password_hash(
                    form.new_password.data
                )
            self.db.session.commit()
        return render_template('uxfab/form.html', form=form, title='Profile')    

    # Admin page
    @root_required
    def admin(self):
        if current_user.role != 'admin':
            abort(HTTPStatus.UNAUTHORIZED)
        users = self.models.User.query.all()
        return render_template('IAM/list.html', items=users, title=f"User list")

    @root_required
    def enabler(self,userid):
        if current_user.role != 'admin':
            abort(HTTPStatus.UNAUTHORIZED)
        user = self.models.User.query.get_or_404(userid)
        user.enabled = True
        self.db.session.commit()
        return redirect('/user/admin')

    @root_required
    def remove_user(self,userid):
        if current_user.role != 'admin':
            abort(HTTPStatus.UNAUTHORIZED)
        user = self.models.User.query.get_or_404(userid)
        self.db.session.delete(user)
        self.db.session.commit()
        return redirect('/user/admin')



from functools import wraps
from flask import current_app, request
from flask_login import current_user
from flask_login.config import EXEMPT_METHODS

def root_required(func):
    """
    If you decorate a view with this, it will ensure that the current user is
    logged in and authenticated before calling the actual view. (If they are
    not, it calls the :attr:`LoginManager.unauthorized` callback.) For
    example::
 
    @app.route('/post')
    @root_required
        def post():
            pass
     
    If there are only certain times you need to require that your user is
    logged in, you can do so with::
     
    if not current_user.id == 0: #.is_authenticated:
        return current_app.login_manager.unauthorized()
     
    ...which is essentially the code that this function adds to your views.
 
    """

    @wraps(func)
    def decorated_view(*args, **kwargs):
        if request.method in EXEMPT_METHODS or current_app.config.get("LOGIN_DISABLED"):
            pass
        elif not current_user.id == 1: #.is_authenticated:
            return current_app.login_manager.unauthorized()
        return current_app.ensure_sync(func)(*args, **kwargs)
 
    return decorated_view

def role_required(role):
    """
    If you decorate a view with this, it will ensure that the current user is
    logged in and authenticated before calling the actual view. (If they are
    not, it calls the :attr:`LoginManager.unauthorized` callback.) For
    example::
 
    @app.route('/post')
    @role_required('admin')
        def post():
            pass
     
    """
    def specified_role_required(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            if request.method in EXEMPT_METHODS or current_app.config.get("LOGIN_DISABLED"):
                pass
            else:
                print('Testing if role assigned')
                role_id = current_app.extensions['IAM'].models.Role.query.filter_by(name=role).first().id
                print(role_id)
                assigned_role = current_app.extensions['IAM'].models.RoleRegistration.query.filter_by(user_id=current_user.id).filter_by(role_id=role_id).first()
                print('Assigned role', assigned_role)
                if not assigned_role: return current_app.login_manager.unauthorized()
            return current_app.ensure_sync(func)(*args, **kwargs)
 
        return decorated_view
    return specified_role_required

def check_user_role(role_name):
    """Checks if the user has a certain role

    Args:
        role_name: str | bool | None
            If str, than role name to check. If True, just requires user to be
            authenticated, if False no role required and if None it should be
            anonymous user

    TODO use this function from within role_required decorator
    """
    if role_name is None:
        return not current_user.is_authenticated
    elif role_name is True:
        return current_user.is_authenticated
    elif role_name is False:
        return True
    elif isinstance(role_name, str):
        if not current_user.is_authenticated: return False
        role = current_app.extensions['IAM'].models.Role.query.filter_by(name=role_name).first()
        if role:
            assigned_role = current_app.extensions['IAM'].models.RoleRegistration.query.filter_by(
                user_id=current_user.id
            ).filter_by(role_id=role.id).first()
            if assigned_role: return True
        return current_user.role == role_name
    else:
        raise TypeError('role_name should be str, bool or None')



from flask_login import UserMixin
import secrets

class IAModels(object):
    def __init__(self, db):
        self.db = db
        self._make_models()
    
    def _make_models(self):
        db = self.db

        # Define classes
        class User(UserMixin, db.Model):
            id = db.Column(db.Integer, primary_key=True)
            username = db.Column(db.String, unique=True, nullable=False)
            email = db.Column(db.String)
            phone = db.Column(db.String)
            address = db.Column(db.String)
            password_hash = db.Column(db.String)
            role = db.Column(db.String)
            role_interest = db.Column(db.Integer, default=0)
            enabled = db.Column(db.Boolean)

            @property
            def data_columns(self):
                return [self.username, self.email, self.role, self.role_interest, self.enabled]

            @property
            def data_headers(self):
                return ('Username', 'Email', 'Role', 'Level', 'Enabled')

            @property
            def admin_actions(self):
                return [
                    (f"/user/enable/{self.id}", 'bi bi-app'),
                    (f"/user/remove/{self.id}", 'bi bi-x-circle')
                ]
        self.User = User

        class Role(db.Model):
            id = db.Column(db.Integer, primary_key=True)
            name = db.Column(db.String, unique=True, nullable=False)
        self.Role = Role

        class RoleRegistration(db.Model):
            id = db.Column(db.Integer, primary_key=True)
            user_id = db.Column(db.Integer, db.ForeignKey("user.id"))
            role_id = db.Column(db.Integer, db.ForeignKey("role.id"))
        self.RoleRegistration = RoleRegistration

        class APIKey(db.Model):
            id = db.Column(db.Integer, primary_key=True)
            key = db.Column(
                db.String, unique=True, nullable=False,
                default= lambda: secrets.token_urlsafe(32) #TODO get size from IAM config
            )
            user_id = db.Column(db.Integer, db.ForeignKey("user.id"))
            expiration = db.Column(db.DateTime)
        self.APIKey = APIKey
